# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
This file is part of tendril
See the COPYING, README, and INSTALL files for more information
"""

import re

from .unitbase import UnitBase
from .unitbase import NumericalUnitBase
from .unitbase import Percentage
from .unitbase import GainBase
from .unitbase import parse_none


class Resistance(NumericalUnitBase):
    _regex_std = re.compile(r"^(?P<numerical>[\d]+\.?[\d]*)\s?(?P<order>[umEKMGT])(?P<residual>[\d]*)$")  # noqa
    _ostrs = ['u', 'm', 'E', 'K', 'M', 'G', 'T']
    _dostr = 'E'


class Capacitance(NumericalUnitBase):
    _regex_std = re.compile(r"^(?P<numerical>[\d]+\.?[\d]*)\s?(?P<order>[pnum]?F?)(?P<residual>[\d]*)$")  # noqa
    _ostrs = ['fF', 'pF', 'nF', 'uF', 'mF', 'F']
    _dostr = 'F'
    _osuffix = 'F'


class Inductance(NumericalUnitBase):
    _regex_std = re.compile(r"^(?P<numerical>[\d]+\.?[\d]*)\s?(?P<order>[pnum]?H?)(?P<residual>[\d]*)$")  # noqa
    _ostrs = ['pH', 'nH', 'uH', 'mH', 'H']
    _dostr = 'H'
    _osuffix = 'H'


class Voltage(NumericalUnitBase):
    _regex_std = re.compile(r"^(?P<numerical>[-+]?[\d]+\.?[\d]*)\s?(?P<order>[fpnumkM]?V?)(?P<residual>[\d]*)$")  # noqa
    _ostrs = ['fV', 'pV', 'nV', 'uV', 'mV', 'V', 'kV', 'MV']
    _dostr = 'V'


class VoltageAC(Voltage):
    pass


class VoltageDC(Voltage):
    pass


class DiodeVoltageDC(VoltageDC):
    pass


class Current(NumericalUnitBase):
    _regex_std = re.compile(r"^(?P<numerical>[-+]?[\d]+\.?[\d]*)\s?(?P<order>[fpnum]?A?)(?P<residual>[\d]*)$")  # noqa
    _ostrs = ['fA', 'pA', 'nA', 'uA', 'mA', 'A']
    _dostr = 'A'


class CurrentAC(Current):
    pass


class CurrentDC(Current):
    pass


class Charge(NumericalUnitBase):
    _regex_std = re.compile(r"^(?P<numerical>[-+]?[\d]+\.?[\d]*)\s?(?P<order>[fpnum]?C?)(?P<residual>[\d]*)$")  # noqa
    _ostrs = ['fC', 'pC', 'nC', 'uC', 'mC', 'C']
    _dostr = 'C'


class VoltageGain(GainBase):
    _regex_std = re.compile(r'^(?P<numerical>[-+]?[\d]+\.?[\d]*)\s?(?P<order>(V/V)?(dB)?)(?P<residual>)$')  # noqa
    _gtype = (Voltage, Voltage)
    _orders = [('', 1), ('V/V', 1), ('dB', lambda x: 10 ** (x/20))]
    _dostr = 'V/V'


class PowerRatio(NumericalUnitBase):
    _regex_std = re.compile(r"^(?P<numerical>[-+]?[\d]+\.?[\d]*)\s?(?P<order>dBm)(?P<residual>)$")  # noqa
    _ostrs = ['dBm']
    _dostr = 'dBm'

    def __repr__(self):
        return str(self._value) + self._dostr


class HFE(UnitBase):
    _regex_std = re.compile(r"^(?P<numerical>[\d]+\.?[\d]*)\s?(?P<order>HFE)(?P<residual>)$")  # noqa
    _ostrs = ['HFE']
    _dostr = 'HFE'

    def __repr__(self):
        return str(self._value) + self._dostr


class Continuity(UnitBase):
    _dostr = None
    _parse_func = parse_none

    def __repr__(self):
        return str(self._value)


class DutyCycle(Percentage):
    pass



# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
This file is part of tendril
See the COPYING, README, and INSTALL files for more information
"""

import math

from decimal import Decimal
from .lengths import Length


class CartesianPoint(object):
    unit = 'mm'

    def __init__(self, x, y):
        self.x = self._norm_repr(x)
        self.y = self._norm_repr(y)

    @staticmethod
    def _norm_repr(v):
        return Decimal(v)

    def __eq__(self, other):
        if self.x == other.x and self.y == other.y:
            return True
        else:
            return False


class CartesianLineSegment(object):
    def __init__(self, p1, p2):
        self.p1 = p1
        self.p2 = p2
        if p1 == p2:
            raise Exception("Line can't have zero length")
        if p1.unit != p2.unit:
            raise ValueError

    def x(self, t):
        return self.p1.x + t * (self.p2.x - self.p1.x)

    def y(self, t):
        return self.p1.y + t * (self.p2.y - self.p1.y)

    def t_x(self, x):
        if self.p1.x != self.p2.x:
            return (Decimal(x) - self.p1.x) / (self.p2.x - self.p1.x)
        else:
            raise ZeroDivisionError

    def t_y(self, y):
        if self.p1.y != self.p2.y:
            return (Decimal(y) - self.p1.y) / (self.p2.y - self.p1.y)
        else:
            raise ZeroDivisionError

    def length(self):
        return Length(str(
            math.sqrt((self.p2.x - self.p1.x)**2 + (self.p2.y - self.p1.y)**2)
        ) + self.p1.unit)

    def __contains__(self, p):
        try:
            t = self.t_x(p.x)
            if 0 <= t <= 1:
                try:
                    if t == self.t_y(p.y):
                        return True
                    else:
                        return False
                except ZeroDivisionError:
                    if p.y == self.p1.y:
                        return True
            else:
                return False
        except ZeroDivisionError:
            t = self.t_y(p.y)
            if 0 <= t <= 1 and p.x == self.p1.x:
                return True
            else:
                return False





from tendril.utils.config import ConfigOption
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)

depends = ['tendril.config.core']


config_elements_currency = [
    ConfigOption(
        'BASE_CURRENCY',
        "'INR'",
        "The code for the base currency."
    ),
    ConfigOption(
        'BASE_CURRENCY_SYMBOL',
        "'INR '",
        "The symbol for the base currency."
    ),
    ConfigOption(
        'BASE_CURRENCY_NAME',
        "'Indian Rupees'",
        "The name of the base currency when used with num2words.",
    ),
    ConfigOption(
        'BASE_CURRENCY_LANG',
        "'en_IN'",
        "The language to use for the base currency when used with num2words."
    ),
    ConfigOption(
        "FIXER_IO_API_KEY",
        "",
        "API Key to use with fixer.io for currency rates."
    ),
]


def load(manager):
    logger.debug("Loading {0}".format(__name__))
    manager.load_elements(config_elements_currency,
                          doc="Native Currency and Exchange Rate API Configuration")



# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Currency Types (:mod:`tendril.utils.types.currency`)
====================================================

The :mod:`tendril.utils.types.currency` contains classes which allow for easy
use and manipulation of currency values. The primary focus is on the primary
use cases of Currencies within tendril, i.e. :

- Handling foreign exchange conversions and exchange rates in application code
  without too much fuss.
- Handling currency arithmetic and comparisons.

This module uses a specific `Base Currency`, defined by
:const:`tendril.config.BASE_CURRENCY` and
:const:`tendril.config.BASE_CURRENCY_SYMBOL` and available as this module's
:data:`native_currency_defn` module variable. In case this module is
to be used independent of Tendril, at least those configuration options
*must* be defined in :mod:`tendril.config`.

.. rubric:: Module Contents

.. autosummary::

    native_currency_defn
    CurrencyDefinition
    CurrencyValue

.. seealso:: :mod:`tendril.utils.types`, for an overview applicable to most
             types defined in Tendril.

.. todo:: The core numbers in this module need to switched to
          :class:`decimal.Decimal`.

"""

from cachecontrol.heuristics import ExpiresAfter

from tendril.config import BASE_CURRENCY
from tendril.config import BASE_CURRENCY_SYMBOL
from tendril.config import BASE_CURRENCY_NAME
from tendril.config import BASE_CURRENCY_LANG
from tendril.config import FIXER_IO_API_KEY

from tendril.utils import www
from decimal import Decimal
from numbers import Number
from num2words import num2words

from .unitbase import NumericalUnitBase
from .unitbase import Percentage

# TODO Switch to a heuristic which uses a cached value only if a fresh
# one is not available.
requests_session = www.get_session(heuristic=ExpiresAfter(days=5))


class CurrencyDefinition(object):
    """
    Instances of this class define a currency.

    The minimal requirement to define a currency is a :attr:`code`, which
    would usually be a standard internationally recognized currency code.

    In addition to the :attr:`code`, a currency definition also includes
    an optional :attr:`symbol`, which is used to create string representations
    of currency values in that currency. In the absence of a :attr:`symbol`,
    the :attr:`code` is used in it's place.

    Unless otherwise specified during the instantiation of the class,
    the exchange rate is obtained from internet services by the
    :meth:`_get_exchval` method.

    :param code: Standard currency code.
    :param symbol: Symbol to use to represent the currency. Optional.
    :param exchval: Exchange rate to use, if not automatic. Optional.

    """
    def __init__(self, code, symbol=None, name=None, exchval=None, lang='en'):
        self._code = code
        self._symbol = symbol
        self._name = name
        self._lang = lang
        if exchval is not None:
            self._exchval = exchval
        else:
            self._exchval = self._get_exchval(self._code)

    @property
    def code(self):
        """

        :return: The currency code.
        :rtype: str

        """
        return self._code

    @property
    def symbol(self):
        """

        :return: The currency symbol, or code if no symbol.
        :rtype: str

        """
        if self._symbol is not None:
            return self._symbol
        else:
            return self._code

    @property
    def name(self):
        if self._name is not None:
            return self._name
        else:
            return self._code

    @property
    def lang(self):
        return self._lang or 'en'

    @property
    def exchval(self):
        """
        :return: The exchange rate
        :rtype: float

        """
        return self._exchval

    @property
    def exch_rate(self):
        """
        The exchange rate in a human-friendly string.
        """
        return "{0}1/{1}{2:,.2f}".format(self.symbol, BASE_CURRENCY_SYMBOL, self.exchval)

    @staticmethod
    def _get_exchval(code):
        """
        Obtains the exchange rate of the currency definition's :attr:`code`
        using the `<http://fixer.io>`_ JSON API. The native currency is used
        as the reference.

        :param code: The currency code for which the exchange rate is needed.
        :type code: str
        :return: The exchange rate of currency specified by code vs the native
                 currency.
        :rtype: float

        """

        if BASE_CURRENCY == code:
            return 1
        if code == 'USD':
            return 72
        apiurl = 'http://data.fixer.io/api/latest?'
        params = {'access_key': FIXER_IO_API_KEY,
                  'symbols': ','.join([code, BASE_CURRENCY])}
        r = requests_session.get(apiurl, params=params)
        data = r.json()
        try:
            rate = float(data['rates'][BASE_CURRENCY]) / float(data['rates'][code])
        except KeyError:
            raise
        return rate

    def __eq__(self, other):
        """
        Two instances of :class:`CurrencyDefinition` will evaluate to be equal
        only when all three parameters of the instances are equal.
        """
        if self.code != other.code:
            return False
        if self.symbol != other.symbol:
            return False
        if self.exchval != other.exchval:
            return False
        return True

    def __repr__(self):
        return "<CurrencyDefinition {0} {1} {2}>" \
               "".format(self.code, self.symbol, self.exchval)


#: The native currency definition used by the module
#:
#: This definition uses the code contained in
#: :const:`tendril.config.BASE_CURRENCY` and symbol
#: :const:`tendril.config.BASE_CURRENCY_SYMBOL`. Application
#: code should import this definition instead of creating new currency
#: definitions whenever one is needed to represent a native currency value.
native_currency_defn = CurrencyDefinition(code=BASE_CURRENCY, 
                                          symbol=BASE_CURRENCY_SYMBOL,
                                          name=BASE_CURRENCY_NAME,
                                          lang=BASE_CURRENCY_LANG)


def _currency_parser(cv, value):
    # TODO Also manage currency definition?
    parts = value.split(' ')
    if len(parts) > 1:
        assert parts[0].strip() in [cv._currency_def.code,
                                    cv._currency_def.symbol]
        return Decimal(parts[1])
    else:
        return Decimal(parts[0])


class CurrencyValue(NumericalUnitBase):
    """
    Instances of this class define a specific currency value, or a certain
    sum of money.

    The `currency_def` can either be a :class:`CurrencyDefinition` instance
    (recommended), or a string containing the code for the currency.

    :param val: The numerical value.
    :param currency_def: The currency definition within which the value
                         is defined.
    :type currency_def: :class:`CurrencyDefinition` or str

    .. note:: Since the exchange rate is obtained at the instantiation of
              the :class:`CurrencyDefinition`, using a string instead of a
              predefined :class:`CurrencyDefinition` instance may result in
              instances of the same currency, but with different exchange
              rates.

    :ivar _currency_def: The currency definition of the source value of the
                         instance.
    :ivar _val: The numerical value in the source currency of the instance.

    .. rubric:: Arithmetic Operations

    .. autosummary::

        __add__
        __sub__
        __mul__
        __div__
        _cmpkey

    """

    _has_bare_order = True
    _orders = [('', 1)]
    _parse_func = _currency_parser

    def __init__(self, value, currency_def=native_currency_defn):
        if isinstance(currency_def, CurrencyDefinition):
            self._currency_def = currency_def
        else:
            self._currency_def = CurrencyDefinition(currency_def)
        super(CurrencyValue, self).__init__(value)

    @property
    def native_value(self):
        """
        The numerical value of the currency value in the native currency,
        i.e., that defined by :data:`native_currency_defn`.

        :rtype: float

        """
        return self._value * self._currency_def.exchval

    @property
    def native_string(self):
        """
        The string representation of the currency value in the native
        currency, i.e., that defined by :data:`native_currency_defn`.

        :rtype: str

        """
        return "{0} {1:,.2f}".format(BASE_CURRENCY_SYMBOL, self.native_value)
    
    @property
    def native_words(self):
        try:
            return num2words(self.native_value, to='currency', adjective=True,
                             currency=native_currency_defn.code,
                             lang=native_currency_defn.lang).title()
        except NotImplementedError:
            return "{0} {1}".format(
                native_currency_defn.name,
                num2words(self.native_value,
                          lang=native_currency_defn.lang).title()
            )

    @property
    def source_value(self):
        """
        The numerical value of the currency value in the source currency,
        i.e., that defined by :attr:`source_currency`.

        :rtype: float

        """
        return self._value

    @property
    def source_string(self):
        """
        The string representation of the currency value in the source
        currency, i.e., that defined by :attr:`source_currency`.

        :rtype: str

        """
        return "{0} {1:,.2f}".format(self._currency_def.symbol, self._val)
    
    @property
    def source_words(self):
        try:
            return num2words(self.source_value, to='currency', adjective=True,
                             currency=self._currency_def.code,
                             lang=self._currency_def.lang).title()
        except NotImplementedError:
            return "{0} {1}".format(
                self._currency_def.name,
                num2words(self.source_value,
                          lang=self._currency_def.lang).title()
            )

    @property
    def source_currency(self):
        """
        The currency definition of the source currency, i.e, the instance
        variable :data:`_currency_def`.

        :rtype: :class:`CurrencyDefinition`

        """
        return self._currency_def

    @property
    def exch_rate(self):
        """
        The applicable exchange rate in a human-friendly string.
        """
        return self._currency_def.exch_rate

    @property
    def is_foreign(self):
        """
        Whether the source currency is Foreign (``True``) or is the native
        currency (``False``).
        """
        if self._currency_def.code != BASE_CURRENCY:
            return True
        else:
            return False

    def __repr__(self):
        return self.native_string

    def __float__(self):
        return float(self.native_value)

    def __add__(self, other):
        """
        Addition of two :class:`CurrencyValue` instances returns a
        :class:`CurrencyValue` instance with the sum of the two operands,
        with currency conversion applied if necessary.

        If the :attr:`source_currency` of the two operands are equal,
        the result uses the the same :attr:`source_currency`. If not,
        the result is uses the :data:`native_currency_defn` as it's
        :attr:`source_currency`.

        If the other operand is a numerical type and evaluates to 0, this
        object is simply returned unchanged.

        Addition with all other Types / Classes is not supported.

        :rtype: :class:`CurrencyValue`
        """
        if isinstance(other, Number) and other == 0:
            return self
        if not isinstance(other, CurrencyValue):
            raise NotImplementedError
        if self._currency_def.symbol == other.source_currency.symbol:
            return CurrencyValue(
                self.source_value + other.source_value, self.source_currency
            )
        else:
            return CurrencyValue(
                self.native_value + other.native_value, native_currency_defn
            )

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __mul__(self, other):
        """
        Multiplication of one :class:`CurrencyValue` instance with a
        numerical type results in a :class:`CurrencyValue` instance,
        whose value is is the currency type operand's value multiplied
        by the numerical operand's value.

        The :attr:`source_currency` of the returned :class:`CurrencyValue`
        is the same as that of the currency type operand.

        Multiplication with all other Types / Classes is not supported.

        :rtype: :class:`CurrencyValue`
        """
        if isinstance(other, Number):
            return CurrencyValue(
                self.source_value * other, self.source_currency
            )
        elif isinstance(other, Percentage):
            return CurrencyValue(
                self.source_value * other.value / 100,
                self.source_currency
            )
        else:
            raise NotImplementedError

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        """
        Division of one :class:`CurrencyValue` instance with a numerical type
        results in a :class:`CurrencyValue` instance, whose value is is the
        currency type operand's value divided by the numerical operand's
        value.

        The :attr:`source_currency` of the returned :class:`CurrencyValue`
        is the same as that of the currency type operand. In this case, the
        first operand must be a :class:`CurrencyValue`, and not the reverse.

        Division of one :class:`CurrencyValue` instance by another returns a
        numerical value, which is obtained by performing the division with the
        operands' :attr:`native_value`.

        Division with all other Types / Classes is not supported.

        :rtype: :class:`CurrencyValue`
        """
        if isinstance(other, Number):
            return CurrencyValue(
                self.source_value / other, self.source_currency
            )
        elif isinstance(other, CurrencyValue):
            return self.native_value / other.native_value
        else:
            raise NotImplementedError

    def __truediv__(self, other):
        return self.__div__(other)

    def __sub__(self, other):
        """
        Subtraction of two :class:`CurrencyValue` instances returns a
        :class:`CurrencyValue` instance with the difference of the two
        operands, with currency conversion applied if necessary.

        If :attr:`source_currency` of the two operands are equal,
        the result uses the the same :attr:`source_currency`. If not,
        the result is in the :data:`native_currency_defn`.

        If the other operand is a numerical type and evaluates to 0, this
        object is simply returned unchanged.

        Subtraction with all other Types / Classes is not supported.

        :rtype: :class:`CurrencyValue`
        """
        if isinstance(other, Number) and other == 0:
            return self
        elif not isinstance(other, CurrencyValue):
            raise NotImplementedError
        else:
            return self.__add__(other.__mul__(-1))

    def _cmpkey(self):
        """
        The comparison of two :class:`CurrencyValue` instances behaves
        identically to the comparison of the operands' :attr:`native_value`.

        Comparison with all other Types / Classes is not supported.
        """
        return self.native_value



# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
This file is part of tendril
See the COPYING, README, and INSTALL files for more information
"""

import re
from decimal import Decimal
from tendril.utils import log
from .unitbase import NumericalUnitBase

logger = log.get_logger(__name__, log.DEFAULT)


lrex = re.compile(r'^((?P<m>[-+]?\d*\.*\d+) *(m|mtr|meter|metre) *$)|((?P<um>[-+]?\d*\.*\d+) *um *$)|((?P<cm>[-+]?\d*\.*\d+) *cm *$)|((?P<mm>[-+]?\d*\.*\d+) *mm *$)|((?P<mil>[-+]?\d*\.*\d+) *mil *$)|((?P<cmil>[-+]?\d*\.*\d+) *cmil *$)|((?P<in>[-+]?\d*\.*\d+) *(in|inch|Inch) *$)|((?P<feet>[-+]?\d*\.*\d+) *(ft|feet|Feet) *$)',  # noqa
                  re.IGNORECASE)


def parse_length(value):
    match = lrex.match(value)
    if match is None:
        logger.warning("Length not parsed : " + value)
        raise ValueError(value)

    um = match.group('um')
    if um is not None:
        _olength = Decimal(um) / Decimal('1000000')
        return _olength

    mm = match.group('mm')
    if mm is not None:
        _olength = Decimal(mm) / Decimal('1000')
        return _olength

    cm = match.group('cm')
    if cm is not None:
        _olength = Decimal(cm) / Decimal('100')
        return _olength

    m = match.group('m')
    if m is not None:
        _olength = Decimal(m)
        return _olength

    inch = match.group('in')
    if inch is not None:
        _olength = Decimal(inch) * Decimal('25.4') / Decimal('1000')
        return _olength

    feet = match.group('feet')
    if feet is not None:
        _olength = Decimal(feet) * 12 * Decimal('25.4') / Decimal('1000')
        return _olength

    mil = match.group('mil')
    if mil is not None:
        _olength = Decimal(mil) * Decimal('25.4') / Decimal('1000000')
        return _olength

    cmil = match.group('cmil')
    if cmil is not None:
        _olength = Decimal(cmil) * Decimal('25.4') / Decimal('100000')
        return _olength


class Length(NumericalUnitBase):
    _ostrs = ['um', 'mm', 'm', 'km']
    _dostr = 'm'
    _parse_func = staticmethod(parse_length)



# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Base Unit Types (:mod:`tendril.utils.types.unitbase`)
=====================================================

The Types provided in this module are not intended for direct use. Instead,
they provide reusable primitives which can be sub-classed to provide
functional Types.

Ideally, all Unit classes should derive from the :class:`UnitBase` class
provided here. In practice, only the newer Types follow this inheritance,
while the older ones still need to be migrated to this form.

.. rubric:: Module Contents

.. autosummary::

    UnitBase
    NumericalUnitBase
    DummyUnit
    parse_none
    parse_percent
    Percentage

.. seealso:: :mod:`tendril.utils.types`, for an overview applicable to
             most types defined in Tendril.

"""

import re
import six
import numbers

from math import log10
from math import floor
from decimal import Decimal
from decimal import InvalidOperation
from fractions import Fraction

from . import ParseException


def round_to_n(x, n):
    if x:
        return round(x, -int(floor(log10(x))) + (n - 1))
    return 0


def remove_exponent(num):
    return num.to_integral() if num == num.to_integral() else num.normalize()


class TypedComparisonMixin(object):
    """
    This mixin allows implementing comparison operators in a Python 3
    compatible way.

    Two instances of a class are compared using their :func:`_cmpkey`
    methods. If the instances have a different ``__class__``, the
    comparison is not implemented. A single exception is implemented,
    for when the other instance is of a numerical type, with value 0.
    """
    def _compare(self, other, method):
        if self.__class__ != other.__class__:
            if other is None:
                return method(self._cmpkey(), 0)
            if not isinstance(other, numbers.Number) or other != 0:
                raise TypeError(
                    "Comparison of : " + repr(self) + ", " + repr(other)
                )
            else:
                return method(self._cmpkey(), other)
        return method(self._cmpkey(), other._cmpkey())

    def __lt__(self, other):
        return self._compare(other, lambda s, o: s < o)

    def __le__(self, other):
        return self._compare(other, lambda s, o: s <= o)

    def __eq__(self, other):
        return self._compare(other, lambda s, o: s == o)

    def __ge__(self, other):
        return self._compare(other, lambda s, o: s >= o)

    def __gt__(self, other):
        return self._compare(other, lambda s, o: s > o)

    def __ne__(self, other):
        return self._compare(other, lambda s, o: s != o)

    def _cmpkey(self):
        raise NotImplementedError


class UnitBase(object):
    """
    The base class for all :mod:`tendril.utils.types` units.

    When instantiated, the `value` param is processed as follows :

    - :class:`str` value is passed though `_parse_func`, and whatever it
      returns is stored.
    - :class:`numbers.Number` values are first converted into
      :class:`decimal.Decimal` and then stored.
    - All other `value` types are simply stored as is.

    :param value: The core value to be stored.
    :param _dostr: The canonical unit / order string to use.
    :param _parse_func: The function used to parse string values into an
                        actual value in the canonical unit.

    .. rubric:: Arithmetic Operations

    This class supports no arithmetic operations.

    """
    _dostr = None
    _parse_func = None

    def __init__(self, value):
        if isinstance(value, (six.text_type, six.string_types)):
            try:
                value = self._parse_func(value)
            except Exception as e:
                raise ParseException(value, e)
        if isinstance(value, numbers.Number):
            if isinstance(value, Fraction):
                value = float(value)
            if not isinstance(value, Decimal):
                value = Decimal(value)

        self._value = value

    # def __float__(self):
    #     return float(self._value)
    #
    # def __int__(self):
    #     return int(self._value)

    @property
    def value(self):
        """
        :return: The core value of the Unit instance, in it's canonical unit.
        """
        return self._value

    def __repr__(self):
        return str(self._value) + self._dostr


class NumericalUnitBase(TypedComparisonMixin, UnitBase):
    """
    The base class for all :mod:`tendril.utils.types` numerical units.

    Provides the patterns used by the various Numerical Units to provide
    their functionality. This class represents and implements the core
    ideas that remain valid across Units (for the most part). The various
    methods and functions implemented here establish the minimum required
    functionality and behaviour expected of all numerical units.

    Specific numerical unit classes may override the methods present here
    to tweak the implementation and/or the interface as per the
    requirements of the quantity they represent, as long as they stay true
    to the spirit of the architecture.

    :param value: The core value to be stored.

    .. seealso:: :class:`UnitBase`

    .. rubric:: The `_orders` / `_ostrs` Class Variables

    Supported units can be provided as a list of strings (`_ostrs`) or a
    list of tuples (`_orders`).

    - In case `_ostrs` is provided, it is assumed that each string
      represents a unit value 1000 times smaller than the next.
    - In case `_orders` is provided, it is assumed that the first
      element of each tuple is the string representation of the order,
      and the second element is the multiplicative factor relative to
      the default order string.
    - In both cases, note that first order within which the unit value's
      representation lies between 1 and 1000 is used to produce the unit's
      string representation. As such, you should place higher priority or
      more 'standard' units towards the beginning of the list.

    .. rubric:: Arithmetic Operations

    .. autosummary::

        __add__
        __sub__
        __mul__
        __div__
        __cmpkey

    """
    _orders = None
    _ostrs = None
    _osuffix = None

    _parse_func = None
    _regex_std = None
    _allow_nOr = True

    _order_type = None
    _order_dict = None
    _has_bare_order = False

    _rorders = None
    _pluralize_ostr = False
    _separate_unit = False

    def __init__(self, value):
        if not self._orders:
            doidx = self._ostrs.index(self._dostr)
            self._orders = [(ostr, (3 * (idx - doidx)))
                            for idx, ostr in enumerate(self._ostrs)]
            self._order_type = 0
        elif not self._ostrs:
            self._ostrs = [order[0] for order in self._orders]
            self._order_type = 1
        if not self._order_dict:
            self._order_dict = {o[0]: o[1] for o in self._orders}
        if self._parse_func is None:
            self._parse_func = self._standard_parser
        if not self._has_bare_order:
            try:
                value = Decimal(value)
            except InvalidOperation:
                pass
        super(NumericalUnitBase, self).__init__(value)

    def _standard_parser(self, value):
        if not self._regex_std:
            raise Exception("Standard parser requires a defined regex")
        match = self._regex_std.match(value)
        n = Decimal(match.group('numerical'))
        r = match.group('residual')
        if r:
            if not self._allow_nOr:
                raise ValueError("nOr structure not allowed for this unit")
            r = Decimal(r)
            n += r/10
        o = match.group('order')
        if self._osuffix and not o.endswith(self._osuffix):
            o += self._osuffix
        try:
            if self._order_type:
                f = self._order_dict[o]
                if isinstance(f, numbers.Number):
                    return n * f
                else:
                    return f(n)
            else:
                exp = self._order_dict[o]
                if exp >= 0:
                    return n * (10 ** exp)
                else:
                    # TODO Significant error in precision here
                    return n / (10 ** abs(exp))
        except KeyError:
            raise ValueError('Order unrecognized : {0} for {1}'
                             ''.format(o, self.__class__))

    def __float__(self):
        return float(self._value)

    def __add__(self, other):
        """
        Addition of two Unit class instances of the same type returns a
        Unit class instance of the same type, with the sum of the two
        operands as it's value.

        If the other operand is a numerical type and evaluates to 0, this
        object is simply returned unchanged.

        Addition with all other Types / Classes is not supported.
        """
        if isinstance(other, numbers.Number) and other == 0:
            return self
        elif self.__class__ == other.__class__:
            return self.__class__(self._value + other._value)
        else:
            return NotImplemented

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __mul__(self, other):
        """
        Multiplication of one Unit type class instance with a numerical type
        results in a Unit type class instance of the same type, whose value
        is the Unit type operand's value multiplied by the numerical operand's
        value.

        Multiplication with all other Types / Classes is not supported.
        """
        if isinstance(other, numbers.Number):
            if isinstance(self, FactorBase):
                return self.__class__(other * self.value)
            if isinstance(other, Decimal):
                return self.__class__(self.value * other)
            return self.__class__(self.value * Decimal(other))
        if isinstance(other, Percentage):
            return self.__class__(self.value * other.value / 100)
        if isinstance(other, GainBase):
            if isinstance(self, GainBase):
                if self._gtype != other._gtype:
                    raise TypeError("Gain is of a different type.")
                return self.__class__(self.value * other.value)
            if other._gtype and not isinstance(self, other._gtype[1]):
                raise TypeError("Gain is of a different type.")
            if isinstance(self, other._gtype[0]):
                return self.__class__(self.value * other.value)
            else:
                return other._gtype[0](self.value * other.value)
        else:
            return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        """
        Division of one Unit type class instance with a numerical type
        results in a Unit type class instance of the same type, whose value
        is the Unit type operand's value divided by the numerical operand's
        value.

        In this case, the first operand must be a Unit type class instance,
        and not the reverse.

        Division of one Unit type class instance by another of the same type
        returns a numerical value, which is obtained by performing the
        division with the operands' value.

        Division with all other Types / Classes is not supported.
        """
        if isinstance(other, numbers.Number):
            if isinstance(other, Decimal):
                return self.__class__(self.value / other)
            else:
                return self.__class__(self.value / Decimal(other))
        elif isinstance(other, Percentage):
            return self.__class__(self.value / other.value * 100)
        elif isinstance(other, self.__class__):
            return self.value / other.value
        else:
            return NotImplemented

    def __rdiv__(self, other):
        return NotImplemented

    def __truediv__(self, other):
        return self.__div__(other)

    def __rtruediv__(self, other):
        return self.__rdiv__(other)

    def __sub__(self, other):
        """
        Subtraction of two Unit class instances of the same type returns a
        Unit class instance of the same type, with the difference of the two
        operands as it's value.

        If the other operand is a numerical type and evaluates to 0, this
        object is simply returned unchanged.

        Subtraction with all other Types / Classes is not supported.
        """
        if isinstance(other, numbers.Number) and other == 0:
            return self
        else:
            return self.__add__(other.__mul__(-1))

    def __rsub__(self, other):
        return other.__sub__(self)

    def __abs__(self):
        if self._value < 0:
            return self.__class__(self._value * -1)
        else:
            return self

    def _cmpkey(self):
        """
        The comparison of two Unit type class instances of the
        same type behaves identically to the comparison of the
        operands' values.

        Comparison with all other Types / Classes is not supported.
        """
        return self._value

    def _pluralize(self, value, ostr):
        if value != 1 and self._pluralize_ostr:
            return value, ostr + 's'
        else:
            return value, ostr

    @property
    def natural_repr(self):
        if self._order_type == 1:
            return self._pluralize(self._value, self._dostr)
        ostr = self._dostr
        value = self._value
        done = False
        while not done:
            ostri = self._ostrs.index(ostr)
            if 1 <= abs(value) < 1000:
                done = True
            elif abs(value) >= 1000:
                if ostri < len(self._ostrs) - 1:
                    ostr = self._ostrs[ostri + 1]
                    value /= Decimal(1000)
                else:
                    done = True
            elif abs(value) < 1:
                if ostri > 0:
                    ostr = self._ostrs[ostri - 1]
                    value *= Decimal(1000)
                else:
                    done = True
        return self._pluralize(value, ostr)

    def _pack_str(self, num, unit):
        if not isinstance(num, six.string_types):
            num = str(num)
        if self._separate_unit:
            return ' '.join([num, unit])
        else:
            return num + unit

    @property
    def quantized_repr(self):
        num, unit = self.natural_repr

        neg = False
        if num < 0:
            neg = True
        num = str(round_to_n(float(abs(num)), 5))
        if neg is True:
            num = '-' + num

        return self._pack_str(num, unit)

    @property
    def integral_repr(self):
        num, unit = self.natural_repr

        neg = False
        if num < 0:
            neg = True
        if num == 0:
            return str(num) + unit
        num = str(round_to_n(float(abs(num)), 2))
        if neg is True:
            num = '-' + num
        return self._pack_str(num, unit)

    @property
    def integral_autoscaled_repr(self):
        value = self._value
        rval = None
        for unit, factor in self._rorders:
            uv, r = divmod(value, factor)
            value = r
            if not uv:
                continue
            part = self._pack_str(*self._pluralize(uv, unit))
            if rval:
                rval = ', '.join([rval, part])
            else:
                rval = part
        return rval

    def fmt_repr(self, fmt):
        num, unit = self.natural_repr
        parts = []
        idx = 0
        if '%v' in fmt:
            fmt = fmt.replace('%v', '{' + str(idx) + '}')
            # TODO Consider applying remove_exponent across the board
            parts.append(remove_exponent(num))
            idx += 1
        if '%u' in fmt:
            fmt = fmt.replace('%u', '{' + str(idx) + '}')
            parts.append(unit)
            idx += 1
        return fmt.format(*parts)

    def __repr__(self):
        if self._rorders:
            return self.integral_autoscaled_repr
        return self._pack_str(*self.natural_repr)


class DummyUnit(UnitBase):
    """
    This class provides a type for placeholder objects. The original
    use case is for handling wave boundaries in streaming protocols.

    Does not support any arithmetic operations.
    """
    def __repr__(self):
        return "Dummy Unit"


def parse_none(value):
    """
    A placeholder parse function which can be used if the Unit
    requires / supports no parsing.
    """
    return value


class FactorBase(NumericalUnitBase):
    pass


class Percentage(FactorBase):
    """
    A base Unit class which provides support for Types that are essentially
    percentages.

    The contribution this base class makes is to be able to parse percentage
    strings so that the Descendant class need not.

    Only the standard :class:`NumericalUnitBase` Arithmetic is supported by
    this class at this time.
    """
    _has_bare_order = True
    _orders = [('%', Decimal('1')), ('pc', Decimal('1')), ('', 100)]
    _dostr = '%'
    _allow_nOr = False
    _regex_std = re.compile(r"^(?P<numerical>[\d]+\.?[\d]*)\s?(?P<order>(pc)?%?)(?P<residual>)$")  # noqa


class Tolerance(Percentage):
    pass


class GainBase(FactorBase):
    _inverse_class = None
    _gtype = None

    def __rdiv__(self, other):
        """
        Division of one Unit type class instance with a numerical type
        results in a Unit type class instance of the same type, whose value
        is the numerical operand's value divided by the unit type operand's
        value.

        In this case, the second operand must be a Unit type class instance,
        and not the reverse.
        """
        if isinstance(other, numbers.Number):
            if not self._inverse_class:
                if not self._gtype[0] == self._gtype[1]:
                    raise TypeError
                inv_cls = self.__class__
            else:
                inv_cls = self._inverse_class
            if isinstance(other, Decimal):
                return inv_cls(other / self.value)
            else:
                return inv_cls(Decimal(other) / self.value)
        else:
            return NotImplemented

    def in_db(self):
        return 20 * log10(self._value)



# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
This file is part of tendril
See the COPYING, README, and INSTALL files for more information
"""

import arrow.arrow

from .unitbase import NumericalUnitBase
from .unitbase import Percentage

from decimal import Decimal
from numbers import Number
import re


class Frequency(NumericalUnitBase):
    _regex_std = re.compile(r"^(?P<numerical>[\d]+\.?[\d]*)\s?(?P<order>[mkMG]?Hz)(?P<residual>)$")  # noqa
    _ostrs = ['mHz', 'Hz', 'kHz', 'MHz', 'GHz']
    _dostr = 'Hz'
    _allow_nOr = False

    def __rdiv__(self, other):
        if isinstance(other, Percentage):
            other = other.value
        if isinstance(other, Number):
            if not isinstance(other, Decimal):
                other = Decimal(other)
            return TimeSpan(other / self._value)
        return NotImplemented

    def __rtruediv__(self, other):
        return self.__rdiv__(other)


class DateSpan(NumericalUnitBase):
    _regex_std = re.compile(r"^(?P<numerical>[-+]?[\d]+\.?[\d]*)\s?(?P<order>(y(ea)?r?)|(d(ay)?))s?(?P<residual>)$")  # noqa
    _orders = [('year', lambda x: x * 365),
               ('yr', lambda x: x * 365),
               ('y', lambda x: x * 365),
               ('day', 1), ('d', 1)]
    _rorders = [('year', 365),
                ('day',  1)]
    _dostr = 'day'
    _allow_nOr = False
    _pluralize_ostr = True
    _separate_unit = True


class TimeSpan(NumericalUnitBase):
    _regex_std = re.compile(r"^(?P<numerical>[-+]?[\d]+\.?[\d]*)\s?(?P<order>[fpnum]?s)(?P<residual>)$")  # noqa
    _ostrs = ['fs', 'ps', 'ns', 'us', 'ms', 's']
    _dostr = 's'
    _allow_nOr = False

    def __init__(self, value):
        if isinstance(value, TimeDelta):
            if value.years != 0 or value.months != 0:
                raise ValueError(
                    'TimeDelta instances used to construct TimeSpan instances'
                    'cannot have non-zero years and months.'
                )
            value = value.microseconds / Decimal('1000000.0') + value.seconds + \
                value.minutes * 60 + value.hours * 3600 + \
                value.days * 3600 * 24
        elif not isinstance(value, Number) and not isinstance(value, str):
            raise TypeError(
                "Can't construct timespan object from {0}".format(value)
            )
        super(TimeSpan, self).__init__(value)

    def __repr__(self):
        if self._value >= 60:
            return repr(self.timedelta)
        else:
            return super(TimeSpan, self).__repr__()

    @property
    def timedelta(self):
        seconds = int(self._value)
        microseconds = int((self._value-seconds) * 1000000)
        return TimeDelta(seconds=seconds, microseconds=microseconds)

    def __rdiv__(self, other):
        if isinstance(other, Percentage):
            other = other.value
        if isinstance(other, Number):
            if not isinstance(other, Decimal):
                other = Decimal(other)
            return Frequency(other / self._value)
        return NotImplemented

    def __rtruediv__(self, other):
        return self.__rdiv__(other)


class TimeStamp(arrow.arrow.Arrow):
    def __sub__(self, other):
        if isinstance(other, TimeDelta):
            return self.clone().replace(years=-other.years,
                                        months=-other.months,
                                        days=-other.days,
                                        hours=-other.hours,
                                        minutes=-other.minutes,
                                        seconds=-other.seconds,
                                        microseconds=-other.microseconds)
        elif isinstance(other, TimeStamp):
            return TimeDelta(years=self.year-other.year,
                             months=self.month-other.month,
                             days=self.day-other.day,
                             hours=self.hour-other.hour,
                             minutes=self.minute-other.minute,
                             seconds=self.second-other.second,
                             microseconds=self.microsecond-other.microsecond)
        else:
            return NotImplemented

    def __add__(self, other):
        if isinstance(other, TimeDelta):
            return self.clone().replace(years=other.years,
                                        months=other.months,
                                        days=other.days,
                                        hours=other.hours,
                                        minutes=other.minutes,
                                        seconds=other.seconds,
                                        microseconds=other.microseconds)
        elif isinstance(other, TimeStamp):
            # return self.clone().replace(years=other.year,
            #                             months=other.month,
            #                             days=other.day,
            #                             hours=other.hour,
            #                             minutes=other.minute,
            #                             seconds=other.second,
            #                             microseconds=other.microsecond)
            raise ValueError
        else:
            return NotImplemented


class TimeDelta(object):
    def __init__(self, years=0, months=0, days=0, hours=0, minutes=0,
                 seconds=0, microseconds=0):
        self.years = years
        self.months = months
        self.days = days
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds
        self.microseconds = microseconds

    def __sub__(self, other):
        if isinstance(other, TimeDelta):
            return TimeDelta(
                years=self.years - other.years,
                months=self.months - other.months,
                days=self.days - other.days,
                hours=self.hours - other.hours,
                minutes=self.minutes - other.minutes,
                seconds=self.seconds - other.seconds,
                microseconds=self.microseconds - other.microseconds
            )
        else:
            return NotImplemented

    def __add__(self, other):
        if isinstance(other, TimeDelta):
            return TimeDelta(
                years=self.years + other.years,
                months=self.months + other.months,
                days=self.days + other.days,
                hours=self.hours + other.hours,
                minutes=self.minutes + other.minutes,
                seconds=self.seconds + other.seconds,
                microseconds=self.microseconds + other.microseconds
            )
        else:
            return NotImplemented

    def __repr__(self):
        return ':'.join(
            [str(self.years), str(self.months), str(self.days),
             str(self.hours), str(self.minutes), str(self.seconds),
             str(self.microseconds)]
        )

    @property
    def timespan(self):
        return TimeSpan(self)


timestamp_factory = arrow.ArrowFactory(TimeStamp)



# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
The Types Package (:mod:`tendril.utils.types`)
==============================================

The :mod:`tendril.utils.types` package, like all the :mod:`tendril.utils`
modules, includes code that ideally resides outside of Tendril itself,
perhaps using some standard or third-party package. In the case of
:mod:`tendril.utils.types`, this package provides various modules and
consequently classes to handle special data types - usually data structures
and units. These use cases are more thoroughly and effectively handled by
various third-party packages, and its existence in Tendril is largely due
to a combination of the *Not Invented Here* syndrome, as well as the
overwhelming number of possible options available, none of which seem to
be ideal drop in replacements for Tendril core's requirements.

While you can use the units and data structures from
:mod:`tendril.utils.types` independent of the rest of Tendril, you probably
should not. The code itself should function, but it lacks the thoroughness
of type-checking or the efficiency of more full fledged implementations. Even
when using Tendril units to handle data generated by one of the Tendril
submodules, I strongly recommend using a full-fledged unit package instead.
If you'd like to use them anyway, well, you have been warned.

Converters from Tendril types to other forms should be reasonably
straightforward to write, and over time some these converters and/or
re-implementations using an established types/units package will hopefully
find their way into the classes defined within :mod:`tendril.utils.types`.

The general principles of design used for the units defined within this
package are as follows :

- If there is an obvious good-fit python package available to provide the
  units or data structures required, use that instead and simply proxy to that
  package within this module. Data-structures are relatively more complex
  beasts, so the remaining principles don't apply to them. They follow more
  of a *use your best judgement* type of development principle.

- Unit Type classes should, as far as possible, derive from
  :class:`unitbase.UnitBase`.

- The fundamental nature of unit instantiation must remain stable and
  consistent within the various units defined. The underlying implementation
  isn't important as long as it provides the functionality required via the
  accessors Tendril expects. More importantly, changing the underlying
  implementation should not change the way application code uses the units.

- In most cases, primary unit instantiation is using a string containing the
  value the object should hold as well as the unit. Units can also be
  instantiated using a :class:`numbers.Number` instance, but application code
  should avoid doing this. This interface is provided largely to support
  arithmetic operations defined between units.

- For each 'simple' unit class, unless provided by an external library, the
  core information should be a number in the "canonical" unit of that class.
  The class should also store the original string.

- Wherever possible, the numbers stored should be :class:`decimal.Decimal`
  instances and not floats. Number of significant digits should ideally be
  tracked, but presently is not.

- The unit classes should support the bare-minimum set of valid operations.
  As far as possible, the simple mathematical operations performed with two
  elements of the same unitclass should produce a valid result (if
  physically meaningful)

- The unit classes should not produce physically meaningless results. This is
  not something that's tested all that much, so it's likely corner cases have
  slipped through the cracks.

- Be unafraid to throw exceptions. If application code misuses the units, the
  application is asking to be penalized.

.. warning::
  No effort is made to have a complete set of units, or even a complete set of
  units of the same dimensionality as the supported units. Addition of new
  types is done on a lazy basis, as are type conversions, inter-unit
  arithmetic, and other ranges. This policy is likely to change only if this
  module is pulled out of Tendril into it's own units library, and maintaining
  that isn't something I'm likely to do alone.

Types
-----

.. toctree::

   tendril.utils.types.currency
   tendril.utils.types.lengths
   tendril.utils.types.unitbase
   tendril.utils.types.electromagnetic
   tendril.utils.types.thermodynamic
   tendril.utils.types.time

Data Structures / Composite Types
---------------------------------

.. toctree::

   tendril.utils.types.signalbase
   tendril.utils.types.cartesian

.. rubric:: Inheritance Diagram

.. inheritance-diagram::
   tendril.utils.types
   tendril.utils.types.cartesian
   tendril.utils.types.currency
   tendril.utils.types.electromagnetic
   tendril.utils.types.thermodynamic
   tendril.utils.types.lengths
   tendril.utils.types.time
   tendril.utils.types.signalbase
   tendril.utils.types.unitbase

"""


__version__ = '0.1.2'


class ParseException(Exception):
    pass



# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import re
from decimal import Decimal as D
from .unitbase import NumericalUnitBase


class Temperature(NumericalUnitBase):
    _regex_std = re.compile(r'^(?P<numerical>[-+]?[\d]+\.?[\d]*)\s?(?P<order>(mK)?[CFK]?)(?P<residual>)$')  # noqa
    _orders = [('C', lambda x: x + D('273.14')),
               ('F', lambda x: ((x - D('32')) * D('5')) / D('9') + D('273.14')),
               ('K', 1),
               ('mK', D('0.001'))]
    _dostr = 'K'
    _allow_nOr = False

    def __repr__(self):
        return str(self._value) + self._dostr


class ThermalDissipation(NumericalUnitBase):
    _regex_std = re.compile(r'^(?P<numerical>[-+]?[\d]+\.?[\d]*)\s?(?P<order>[numkM]?W)(?P<residual>)$')  # noqa
    _ostrs = ['nW', 'uW', 'mW', 'W', 'kW', 'MW']
    _dostr = 'W'
    _allow_nOr = False



# Copyright (C) 2015 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from .time import timestamp_factory
from .time import TimeStamp
from .time import TimeSpan
from .time import TimeDelta

from .unitbase import Percentage

from collections import deque
import copy


class SignalBase(object):
    def __init__(self, unitclass):
        self._unitclass = unitclass

    @property
    def unitclass(self):
        return self._unitclass


class SignalErrorBar(object):
    def __init__(self, quantization=None, noise_floor=None,
                 error_pc=None):
        self.quantization = quantization
        self.noise_floor = noise_floor
        self.error_pc = error_pc


class SignalPoint(SignalBase):
    def __init__(self, unitclass, value, ts=None):
        super(SignalPoint, self).__init__(unitclass)

        if not isinstance(value, unitclass):
            self.value = unitclass(value)
        else:
            self.value = value

        if ts is None:
            self.timestamp = timestamp_factory.now()
        else:
            self.timestamp = ts

        self._error_bar = None

    @property
    def error_bar(self):
        return self._error_bar

    @error_bar.setter
    def error_bar(self, value):
        self._error_bar = value

    def __repr__(self):
        return "<SignalPoint at " + repr(self.timestamp) + \
               " :: " + repr(self.value) + " >"


class SignalWave(SignalBase):
    def __init__(self, unitclass, points=None, spacing=None, ts0=None,
                 interpolation="piecewise_linear", buffer_size=None,
                 use_point_ts=True, stabilization_length=5,
                 stabilization_pc='1pc'):
        super(SignalWave, self).__init__(unitclass)

        self._stabilization_length = stabilization_length
        if isinstance(stabilization_pc, Percentage):
            self._stabilization_pc = stabilization_pc
        else:
            self._stabilization_pc = Percentage(stabilization_pc)
        self._buffer_size = buffer_size
        self._use_point_ts = use_point_ts

        if spacing is not None:
            if isinstance(spacing, TimeSpan):
                self._spacing = spacing.timedelta
            elif isinstance(spacing, TimeDelta):
                self._spacing = spacing
            else:
                raise TypeError("spacing must be an instance of TimeDelta or "
                                "TimeSpan. Got " + repr(spacing))

        if ts0 is not None:
            if not isinstance(ts0, TimeStamp):
                raise TypeError("ts0 must be an instance of TimeStamp")
            self._ts0 = ts0
        else:
            self._ts0 = timestamp_factory.now()

        if points and isinstance(points, deque):
            if isinstance(points[0], tuple):
                self._points = points
            else:
                if not self._use_point_ts:
                    span_spacing = self._spacing.timespan
                    self._points = deque(
                        [(ts0 + (span_spacing * idx).timedelta, point)
                         for idx, point in enumerate(points)],
                        maxlen=buffer_size
                    )
                else:
                    self._points = deque([(point.timestamp, point)
                                          for point in points],
                                         maxlen=buffer_size)
        else:
            self._points = deque(maxlen=buffer_size)

        self._interpolation = interpolation

    @property
    def last_timestamp(self):
        if len(self._points) == 0:
            if self._ts0 is not None:
                return self._ts0 - self._spacing
            else:
                return None
        return self._points[-1][0]

    @property
    def stabilization_length(self):
        return self._stabilization_length

    @property
    def is_stable(self):
        lval = self._points[-1][1].value
        for i in range(self.stabilization_length):
            if abs(self._points[-1-i][1].value - lval) \
                    > abs(lval * self._stabilization_pc):
                return False
        return True

    @property
    def stabilized_value(self):
        acc = 0
        for i in range(self.stabilization_length):
            acc += self._points[-1-i][1].value
        return acc / self.stabilization_length

    @property
    def ts0(self):
        return self._ts0

    @property
    def max(self):
        return max(self._points, key=lambda x: x[1].value)[1].value

    @property
    def min(self):
        return min(self._points, key=lambda x: x[1].value)[1].value

    @property
    def spread(self):
        return self.max - self.min

    @property
    def mean(self):
        if self.unitclass == int:
            return int(round(float(sum([x[1].value for x in self._points])) /
                             len(self._points)))
        return sum([x[1].value for x in self._points]) / len(self._points)

    @property
    def latest_point(self):
        return self._points[-1][1].value

    @property
    def points(self):
        return sorted(self._points, key=lambda x: x[0])

    @property
    def spacing(self):
        return TimeSpan(self._spacing)

    @property
    def maxlen(self):
        return self._points.maxlen

    @maxlen.setter
    def maxlen(self, value):
        self._points = deque(self._points, maxlen=value)

    def add_point(self, point, ts=None):
        if point.unitclass != self.unitclass:
            raise TypeError
        if ts is None:
            if self._use_point_ts is False:
                if self.last_timestamp is None:
                    self._ts0 = timestamp_factory.now()
                ts = self.last_timestamp + self._spacing
            elif point.timestamp is not None:
                if self.last_timestamp is None:
                    self._ts0 = point.timestamp
                ts = point.timestamp
        if ts is None:
            raise ValueError
        self._points.append((ts, point))

    def __add__(self, other):
        if isinstance(other, SignalWave):
            if self.unitclass != other.unitclass:
                raise TypeError
            rval = copy.copy(self)
            rval.extend(other._points)
            return rval
        elif isinstance(other, SignalPoint):
            self.add_point(other)
            return self
        else:
            raise NotImplementedError

    def __radd__(self, other):
        if isinstance(other, SignalWave):
            if self.unitclass != other.unitclass:
                raise TypeError
            return self._points.extendleft(reversed(other._points))
        elif isinstance(other, SignalPoint):
            if self.unitclass != other.unitclass:
                raise TypeError
            return self._points.appendleft(other)
        else:
            raise NotImplementedError

    def __len__(self):
        return len(self._points)

    def __getitem__(self, item):
        return self._points.__getitem__(item)

    def __setitem__(self, key, value):
        return self._points.__setitem__(key, value)

    def __delitem__(self, key):
        return self._points.__delitem__(key)

    def __iter__(self):
        return self._points.__iter__()

    def __reversed__(self):
        return self._points.__reversed__()

    def __contains__(self, item):
        return self._points.__contains__(item)

    def clear(self):
        self._ts0 = self.last_timestamp
        self._points.clear()

    def popleft(self, *args, **kwargs):
        return self._points.popleft(*args, **kwargs)

    def pop(self, *args, **kwargs):
        return self._points.pop(*args, **kwargs)

    def append(self, *args, **kwargs):
        return self._points.append(*args, **kwargs)

    def appendleft(self, *args, **kwargs):
        return self._points.appendleft(*args, **kwargs)

    def extend(self, *args, **kwargs):
        return self._points.extend(*args, **kwargs)

    def extendleft(self, *args, **kwargs):
        return self._points.extendleft(*args, **kwargs)
